/*
 * Copyright (c) 2006, John Mettraux, OpenWFE.org
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * . Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.  
 * 
 * . Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * 
 * . Neither the name of the "OpenWFE" nor the names of its contributors may be
 *   used to endorse or promote products derived from this software without
 *   specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: Definitions.java 2990 2006-08-06 20:23:49Z jmettraux $
 */

//
// StaticSyndicatedDispatcher.java
//
// john.mettraux@openwfe.org
//
// generated with 
// jtmpl 1.1.01 2004/05/19 (john.mettraux@openwfe.org)
//

package openwfe.org.syndication.impl.dispatch;

import openwfe.org.Utils;
import openwfe.org.MapUtils;
import openwfe.org.ServiceException;
import openwfe.org.ApplicationContext;
import openwfe.org.xml.XmlUtils;
import openwfe.org.net.SocketService;
import openwfe.org.engine.workitem.WorkItem;
import openwfe.org.engine.workitem.CancelItem;
import openwfe.org.engine.workitem.InFlowItem;
import openwfe.org.engine.workitem.InFlowWorkItem;
import openwfe.org.engine.workitem.WorkItemCoder;
import openwfe.org.engine.dispatch.DispatchingException;
import openwfe.org.engine.dispatch.AbstractWorkItemDispatcher;
import openwfe.org.engine.expressions.FlowExpressionId;
import openwfe.org.engine.participants.Participant;
import openwfe.org.engine.impl.workitem.xml.XmlWorkItemCoder;
import openwfe.org.syndication.Feed;
import openwfe.org.syndication.Content;
import openwfe.org.syndication.Link;
import openwfe.org.syndication.Entry;
import openwfe.org.syndication.Person;
import openwfe.org.syndication.FeedCoder;
import openwfe.org.syndication.impl.Atom10Coder;


/**
 * This dispatcher simply stores its feed in a static XML file;
 * the default syndication 'protocol' currently used is Atom1.0.
 *
 * <p><font size=2>CVS Info :
 * <br>$Author$
 * <br>$Id$ </font>
 *
 * @author john.mettraux@openwfe.org
 */
public class StaticSyndicatedDispatcher

    extends AbstractWorkItemDispatcher

{

    private final static org.apache.log4j.Logger log = org.apache.log4j.Logger
        .getLogger(StaticSyndicatedDispatcher.class.getName());

    //
    // CONSTANTS & co

    /**
     * 'feedCoderClass' indicates to this dispatcher how it should
     * turn its feed into XML (which flavour : RSS, Atom ?).
     */
    public final static String P_FEED_CODER_CLASS
        = "feedCoderClass";

    /**
     * The default feedCoder is the 'atom 1.0' one.
     */
    public final static String DEFAULT_FEED_CODER_CLASS
        = Atom10Coder.class.getName();

    /**
     * The 'link' for the feed.
     */
    public final static String P_LINK
        = "link";

    /**
     * The 'outputDirectory' where feeds are stored.
     */
    public final static String P_OUTPUT_DIRECTORY
        = "outputDirectory";

    /**
     * By default, feeds are stored under "work/engine/feeds/".
     */
    public final static String DEFAULT_OUTPUT_DIRECTORY
        = "work/engine/feeds";

    /**
     * The max number of entries the feed will contain
     * is set with the 'feedMaxSize' parameter.
     */
    public final static String P_FEED_MAX_SIZE
        = "feedMaxSize";

    /**
     * The default 'feedMaxSize' is set to 100 (entries).
     */
    public final static int DEFAULT_FEED_MAX_SIZE
        = 100;

    /**
     * The URI of the OpenWFE project; it's used as the Author URI.
     */
    public final static String OPENWFE_URI
        = "http://www.openwfe.org";

    /**
     * This is the text content for a cancelled workitem.
     */
    public final static String CANCELLED
        = "** cancelled **";

    //
    // FIELDS

    private Feed feed = null;
    private String link = null;
    private String feedFile = null;

    private FeedCoder coder = null;

    //
    // CONSTRUCTORS

    public void init
        (final String serviceName,
         final ApplicationContext context,
         final java.util.Map serviceParams)
    throws
        ServiceException
    {
        super.init(serviceName, context, serviceParams);

        //
        // participant name

        final String participantName = 
            (String)serviceParams.get(Participant.PARTICIPANT_NAME);

        log.debug("init() participantName >"+participantName+"<");

        //
        // determine feed coder

        final String coderClassName = MapUtils.getAsString
            (serviceParams, P_FEED_CODER_CLASS, DEFAULT_FEED_CODER_CLASS);

        try
        {
            final Class clazz = Class.forName(coderClassName);
            this.coder = (FeedCoder)clazz.newInstance();
        }
        catch (final Throwable t)
        {
            throw new ServiceException
                ("failed to instantiate feed coder '"+coderClassName+"'",
                 t);
        }

        log.info
            ("init() using coder of class "+this.coder.getClass().getName());

        //
        // feed max size

        final int feedMaxSize = MapUtils.getAsInt
            (serviceParams, P_FEED_MAX_SIZE, DEFAULT_FEED_MAX_SIZE);

        log.info
            ("init() 'feedMaxSize' set to "+feedMaxSize);

        //
        // create feed

        this.feed = new Feed(feedMaxSize);

        this.feed.setTitle
            ("feed for OpenWFE participant '"+participantName+"'");

        this.link = MapUtils.getAsString
            (serviceParams, P_LINK);

        if ( ! this.link.endsWith("/"))
            this.link += "/";

        this.feed.setLink(new Link(this.link));
            // even if it's null

        this.feed.setId(this.link);

        //
        // work directory

        String workDirectory = MapUtils.getAsString
            (serviceParams, P_OUTPUT_DIRECTORY, DEFAULT_OUTPUT_DIRECTORY);

        workDirectory = 
            context.getApplicationDirectory() + 
            java.io.File.separator +
            workDirectory;

        final java.io.File wdf = new java.io.File(workDirectory);
        wdf.mkdirs();

        this.feedFile = 
            workDirectory + java.io.File.separator + 
            "feed_" + participantName + ".xml";

        log.info("init() feedFile is "+this.feedFile);
    }

    //
    // GETTERS and SETTERS

    public Feed getFeed ()
    {
        return this.feed;
    }

    //
    // METHODS from WorkItemDispatcher

    /**
     * Dispatches a workitem somewhere
     */
    public Object dispatch (final WorkItem wi) 
        throws DispatchingException
    {
        if (wi instanceof CancelItem)
            return cancel((CancelItem)wi);

        //
        // prepare entry

        final Entry entry = buildEntry(wi);

        final InFlowWorkItem ifwi = (InFlowWorkItem)wi;

        // add content

        final Content content = new Content();

        final WorkItemCoder coder = instantiateEncoder();

        content.setType(coder.getMimeType());

        try
        {
            if (coder instanceof XmlWorkItemCoder)
            {
                final org.jdom.Element elt = 
                    ((XmlWorkItemCoder)coder).encodeAsXml(ifwi);

                content.setValue(elt);
            }
            else
            {
                final byte[] bwi = (byte[])coder
                    .encode(ifwi, getContext(), getParams());

                content.setValue(new String(bwi, Utils.getEncoding()));
            }
        }
        catch (final Throwable t)
        {
            throw new DispatchingException
                ("failed to encode the workitem in a feed entry", t);
        }

        entry.getContents().add(content);

        //
        // add it to the feed

        this.addEntry(entry);

        return null;
    }

    //
    // METHODS

    /**
     * Prepares a feed entry for the given InFlowItem (be it an
     * InFlowWorkItem or a CancelItem).
     */
    protected Entry buildEntry (final WorkItem item)
    {
        final InFlowItem ifi = (InFlowItem)item;

        final Entry entry = new Entry();

        entry.setId(toId(ifi));
        entry.setTitle(toTitle(ifi));
        entry.setLink(new Link(this.link));
        entry.setUpdated(new java.util.Date()); // now

        entry.setAuthor(new Person(this.getClass().getName(), OPENWFE_URI));

        // description

        final StringBuffer sbDesc = new StringBuffer();

        sbDesc
            .append("Workitem for flow ")
            .append(ifi.getId().getWorkflowInstanceId());

        final Content description = new Content();

        description.setType("text/plain");
        description.setValue(sbDesc.toString());

        entry.setDescription(description);

        return entry;
    }

    /**
     * Adds the entry to the feed and stores the feed.
     */
    protected void addEntry (final Entry e)
        throws DispatchingException
    {
        this.feed.addEntry(e);
        this.feed.setUpdated(new java.util.Date());
            //
            // feed got updated

        this.storeFeed();
    }

    /**
     * Stores the feed in a work file.
     */
    protected void storeFeed ()
        throws DispatchingException
    {
        try
        {
            XmlUtils.save(this.feedFile, getEncodedFeed());
        }
        catch (final Throwable t)
        {
            log.error
                ("failed to store the feed into "+this.feedFile,
                 t);
            throw new DispatchingException
                ("failed to store the feed into "+this.feedFile,
                 t);
        }
    }

    /**
     * Returns the feed encoded as an XML document. this method
     * has been made public to make tests more convenient; only
     * that reason for now.
     */
    public org.jdom.Document getEncodedFeed ()
    {
        return this.coder.encode(this.feed);
    }

    /**
     * Returns the feedCoder instance that this dispatcher uses.
     */
    public FeedCoder getFeedCoder ()
    {
        return this.coder;
    }

    /**
     * A cancel item removes the corresponding workitem from the feed.
     */
    protected Object cancel (final CancelItem ci)
        throws DispatchingException
    {
        /*
        this.feed.removeEntry(toId(ci));
        this.storeFeed();
        */

        final Entry entry = buildEntry(ci);

        final Content con = new Content(CANCELLED);

        entry.getContents().add(con);

        this.addEntry(entry);

        if (log.isDebugEnabled())
            log.debug("cancel() cancelled entry for "+ci.getId());

        return null;
    }

    /**
     * Generates the entry id for a workitem.
     */
    protected String toId (final InFlowItem wi)
    {
        return this.link + toString(wi.getId());
    }

    /**
     * Generates the entry title for a workitem.
     */
    protected String toTitle (final InFlowItem wi)
    {
        return toString(wi.getId());
    }

    /**
     * Turns a FlowExpressionId into a String.
     * Makes sure it's usable in a feed URI.
     */
    protected String toString (final FlowExpressionId fei)
    {
        //return fei.toParseableString().replaceAll("\\|", "_");

        final StringBuffer sb = new StringBuffer();

        sb
            .append(fei.getWorkflowDefinitionName())
            .append("_")
            .append(fei.getWorkflowDefinitionRevision())
            .append("__")
            .append(fei.getWorkflowInstanceId())
            .append("__")
            .append(fei.getExpressionName())
            .append("_")
            .append(fei.getExpressionId());

        String s = sb.toString();

        s = s.replaceAll(" ", "_");

        return s;
    }

    //
    // STATIC METHODS

}
